{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "slide"
    }
   },
   "source": [
    "# Iterators and Generators"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "slide"
    }
   },
   "source": [
    "## Iterators\n",
    "_**Iterators**_ are objects that produce successive items or values from an associated **_iterable_**.\n",
    "They:\n",
    "- Hold the state (position) of the iteration\n",
    "- Allow looping just once and must be reinitialized to loop again\n",
    "- Implement the ***\\_\\_next\\_\\_*** method that...\n",
    "    - returns the next item in the sequence\n",
    "    - raises the ***StopIteration*** exception if there is nothing to return\n",
    "    - can also be invoked using the ***next(iterable)*** function\n",
    "\n",
    "An **_iterable_** is an object that can be iterated over.\n",
    "- Must be capable of returning an iterator\n",
    "- Must implement the ***\\_\\_iter\\_\\_*** method, callable using the *iter* function"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "slide"
    }
   },
   "source": [
    "#### Simple iterables and iterators examples "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "Lets define an _iterable_ and an _iterator_..."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "class Iterable:\n",
    "    def __iter__(self):\n",
    "        \"\"\"\n",
    "        Called by iter(Iterable())\n",
    "        \"\"\" \n",
    "        return Iterator()\n",
    "\n",
    "class Iterator:\n",
    "    def __init__(self):\n",
    "        self.x = -1\n",
    "        \n",
    "    def __next__(self):\n",
    "        \"\"\"\n",
    "        Called by next(iterator)\n",
    "        \"\"\" \n",
    "        self.x += 1\n",
    "        return self.x"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "... and instantiate them."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "-"
    }
   },
   "outputs": [],
   "source": [
    "iterable = Iterable()\n",
    "\n",
    "print(iterable)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "iterator = iter(iterable)\n",
    "# iter(iterable) ==> iterable.__iter__()\n",
    "\n",
    "print(iterator)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Let's call the _next()_ function on the iterator."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "# iterator.__next__() \n",
    "print(next(iterator))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "An object can also define both _next_ and _iter_ methods."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "outputs": [],
   "source": [
    "class SimpleIterable:\n",
    "    def __iter__(self):\n",
    "        self.x = -1\n",
    "        return self\n",
    "\n",
    "#     def __next__(self):\n",
    "#         self.x += 1\n",
    "#         return self.x\n",
    "    \n",
    "    def __next__(self):\n",
    "        if self.x <= 3:\n",
    "            self.x += 1\n",
    "        else:\n",
    "            raise StopIteration\n",
    "        return self.x"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "iterable = SimpleIterable()\n",
    "print(type(iterable))\n",
    "# iterable.x"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "iterator = iter(iterable)\n",
    "type(iterator)\n",
    "# iterable.x"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "# Call next 5 times\n",
    "next(iterator)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "slide"
    }
   },
   "source": [
    "## How to iterate on iterators"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "#### Iterators from containers (e.g. lists)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "a = list(range(4))\n",
    "print(a)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Calling _next()_ on a list won't work."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "next(a)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "But getting an _iterator_ from a list and iterating on it will."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "a = iter(a)\n",
    "# iter(iter) = iter\n",
    "\n",
    "next(a)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "#### The foreach construct\n",
    "- Built-in in the language with the _**for** ... **in**_ construct\n",
    "- It allows looping on all elements of an iterable.\n",
    "- Automatically calls the _iter(...)_ function before starting looping"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "print(\"With iter()\")\n",
    "iterator = iter(iterable)\n",
    "for item in iterator:\n",
    "    print(item)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "print(\"Without iter()\")\n",
    "for item in SimpleIterable():\n",
    "    print(item)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "slide"
    }
   },
   "source": [
    "## Generators\n",
    "- Functions containing the keyword ***yield***\n",
    "- _yield_ :\n",
    "    - works similarly to _return_ and returns an object when called...\n",
    "    - ... but **state of the function is saved**\n",
    "\n",
    "- When _next()_ is called again on the generator function, execution resumes where it was left off\n",
    "\n",
    "- Note that generators **do not return** values when initialized."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "### Examples"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "#### Trivial generator"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "def f():\n",
    "  print(\"-- start --\")\n",
    "  yield 3\n",
    "\n",
    "  print(\"-- middle --\")\n",
    "  yield 4\n",
    "\n",
    "  print(\"-- finished --\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "generator = f()\n",
    "generator"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "next(generator)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "#### Counter"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "def counter():\n",
    "    x : int = 0\n",
    "        \n",
    "    while True:\n",
    "        yield x\n",
    "        x += 1\n",
    "        \n",
    "generator = counter()\n",
    "generator"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "next(generator)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "#### Generators can also be defined inline!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "generator = (x for x in range(10))\n",
    "print(\"generator type:\", type(generator))\n",
    "\n",
    "list(generator)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "source": [
    "#### **Mind the difference with list comprehensions!**"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "not_a_generator = [x for x in range(10)]\n",
    "print(\"not_a_generator type:\", type(not_a_generator))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "slide"
    }
   },
   "source": [
    "### Generators are **lazy iterators**\n",
    "- They are used to **generate values dynamically**\n",
    "    - Very useful to cope, for instance, with Out of Memory issues\n",
    "- As iterators, they don't implement the *\\_\\_len\\_\\_* method\n",
    "    - i.e.,  *len()* function will cause an exception\n",
    "- Generators support **bidirectional communication**.\n",
    "    - You can pass values to the generator **after** its initialization\n",
    "- **Concurrent** and **recursive invocations** are **allowed**...\n",
    "    - ... even though they are **not thread safe** out of the box."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "### Dynamic value generation example"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "-"
    }
   },
   "outputs": [],
   "source": [
    "from datetime import datetime\n",
    "print(\"MS output format:\", datetime.now().microsecond)\n",
    "\n",
    "def very_unsafe_prng(max_value):    \n",
    "    while True:\n",
    "        yield datetime.now().microsecond % max_value\n",
    "        \n",
    "generator = very_unsafe_prng(10)\n",
    "generator"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "# No len!\n",
    "len(generator)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "next(generator)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "### Bidirectional communication\n",
    "Bidirectional communication allows to send values to the generator. Relies on three methods:\n",
    "- ___.send(...)___: sends the value to the generator and, like __next()__, returns the next value\n",
    "- ___.throw(...)___: throws the passed exception after resuming the generator that will handle it\n",
    "- ___.close()___: stops the generator. Equivalent to ___.throw(GeneratorExit())___\n",
    "- _yield_ can be used in expressions to assign values to generator’s variables\n",
    "    - Values will be assigned when the generator resumes from _yield_"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "####  Example"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "outputs": [],
   "source": [
    "from random import choice\n",
    "\n",
    "# Define allowed values ([1, 7])\n",
    "values = list(range(1, 8))\n",
    "\n",
    "# Define a generator\n",
    "def seven_and_half():\n",
    "    values_sum = 0\n",
    "    results = []\n",
    "\n",
    "    player = 1\n",
    "    \n",
    "    # Keep going until generator is closed\n",
    "    while True:\n",
    "        # Pick a card (pseudo) randomly\n",
    "        value = choice(values)\n",
    "\n",
    "        # Accumulate the values\n",
    "        values_sum += value\n",
    "        \n",
    "        # \n",
    "        try:\n",
    "            response = yield value, values_sum\n",
    "        except GeneratorExit:\n",
    "            results.append((player, values_sum),)\n",
    "            \n",
    "            for player, score in results:\n",
    "                print(\"Player {} scored {}\".format(player, score))\n",
    "            break\n",
    "        \n",
    "        if response is False or response is None:            \n",
    "            results.append((player, values_sum),)\n",
    "            values_sum = 0\n",
    "            player += 1\n",
    "            \n",
    "    print(\"Exiting\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "outputs": [],
   "source": [
    "# Init the generator\n",
    "generator = seven_and_half()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "keep_playing = None\n",
    "\n",
    "# Keep in mind that\n",
    "# generator.send(None) === next(generator)\n",
    "    \n",
    "while True:  \n",
    "    if keep_playing is False:\n",
    "        break\n",
    "        \n",
    "    print(\"Picking a card.\")\n",
    "    value, values_sum = generator.send(keep_playing)\n",
    "        \n",
    "    print(\"Picked {}. Total: {}\".format(value, values_sum))\n",
    "    \n",
    "    if values_sum > 7:\n",
    "        print(\"You lost!\")\n",
    "        keep_playing = False    \n",
    "    else:\n",
    "        keep_playing : bool = (input(\"Keep picking? \") in [\"y\", \"Y\", True] )\n",
    "    # print(output)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "generator.close()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "slide"
    }
   },
   "source": [
    "## Exercise"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "#### Intro\n",
    "CSV (Comma-Separated Values) files are text files where **each row is a** data **record** and **columns** are **separated by commas** (or some other character).\n",
    "\n",
    "The first row is (usually) the header (i.e., the name of the corresponding column).\n",
    "\n",
    "A CSV file looks like:\n",
    "\n",
    "    id,name,surname\n",
    "    0,Mickey,Mouse"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "#### Request\n",
    "\n",
    "You have to **process a _CSV_ file**. Lets assume it is **too large to fit in RAM**.\n",
    "\n",
    "You should process it in small pieces, e.g., by reading each line sequentially using a generator.\n",
    "\n",
    "Specifically, after reading the header, for each line of the file create a dictionary with the column-value associations."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "source": [
    "#### A few tips\n",
    "- *Pathlib* module offers classes representing filesystem paths with semantics appropriate for different operating systems\n",
    "- Use the _**zip**(*iterables)_ builtin function. [From the docs](https://docs.python.org/3/library/functions.html#zip):\n",
    "    - Builds an iterator that aggregates elements from each of the iterables\n",
    "    - That is, it returns an iterator of tuples, i.e., The i-th element of the tuple contains the i-th element from each of the argument sequences or iterables.\n",
    "    - The iterator stops when the shortest input iterable is exhausted\n",
    "- Use the ***with*** statement. [From the docs](https://docs.python.org/3/reference/compound_stmts.html#with): the with statement is used to wrap the execution of a block with methods defined by a context manager. This allows common *try…except…finally* usage patterns to be encapsulated for convenient reuse.\n",
    "- As an example CSV, download [*as raw*](https://raw.githubusercontent.com/Currie32/500-Greatest-Albums/master/albumlist.csv) the [Rolling Stone Magazine's list of \"The 500 Greatest Albums of All Time.\"](https://github.com/Currie32/500-Greatest-Albums/blob/master/albumlist.csv) from GitHub."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Solution"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "subslide"
    }
   },
   "outputs": [],
   "source": [
    "from pathlib2 import Path\n",
    "\n",
    "\n",
    "def dataset_reader(file):\n",
    "    # Lets use pathlib instead of using the open() function,\n",
    "    # with open(file, \"r+\") as f:\n",
    "\n",
    "    # Creating a Path instance.\n",
    "    file = Path(file)\n",
    "\n",
    "    # Not actually needed, just showing some functionality\n",
    "    if not file.absolute():\n",
    "        file = file.resolve()\n",
    "        \n",
    "    print(\"Opening\", file.name, \"in folder\", file.parent)\n",
    "    \n",
    "    if not file.exists():\n",
    "        raise FileNotFoundError(\"File doesn't exist!\")\n",
    "        \n",
    "    \n",
    "    with file.open(\"r+\", encoding=\"ISO-8859-15\") as f:\n",
    "        header = f.readline()\n",
    "        columns = header.strip().split(',')\n",
    "        \n",
    "        print(\"Found columns:\", columns)\n",
    "        for line in f:\n",
    "            values = line.strip().split(',')\n",
    "#             print(values)\n",
    "            \n",
    "            try: \n",
    "                yield dict(zip(columns, values))\n",
    "            except GeneratorExit:\n",
    "                print(\"Closing the generator!\")\n",
    "                break"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "file = \"albumlist.csv\"\n",
    "\n",
    "generator = dataset_reader(file)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "slideshow": {
     "slide_type": "fragment"
    }
   },
   "outputs": [],
   "source": [
    "next(generator)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "slideshow": {
     "slide_type": "slide"
    }
   },
   "source": [
    "---\n",
    "## References\n",
    "\n",
    "* [Scipy Lectures](https://scipy-lectures.org/advanced/advanced_python/index.html)\n",
    "* [StackAbuse](https://stackabuse.com/introduction-to-python-iterators/)\n",
    "* [RealPython](https://realpython.com/introduction-to-python-generators/)"
   ]
  }
 ],
 "metadata": {
  "celltoolbar": "Slideshow",
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.6"
  },
  "rise": {
   "enable_chalkboard": true,
   "footer": "<h3 class='text-muted'></h3>",
   "header": "<h3 class='text-muted'>Advanced Programming Languages: <i>python</i> - marco.grassia@unict.it</h3>",
   "scroll": true,
   "start_slideshow_at": "selected",
   "theme": "night"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
